package it.inaf.oats.ia2.tapschemamanager.datalayer;

import java.io.Closeable;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

class SchemaInfo implements Serializable {

    private static final long serialVersionUID = 2835388936102349727L;

    private final List<String> allSchemaTables;
    private final List<KeyEntity> schemaKeys;
    private final Map<String, String> tablesType;

    SchemaInfo(Connection sourceConnection, String schemaName) throws SQLException {
        allSchemaTables = DataProvider.getAllTablesNames(sourceConnection, schemaName);
        schemaKeys = DataProvider.getSchemaKeys(sourceConnection, schemaName);
        tablesType = DataProvider.getTablesTypes(sourceConnection, schemaName);
    }

    public List<String> getAllSchemaTables() {
        return allSchemaTables;
    }

    public List<KeyEntity> getSchemaKeys() {
        return schemaKeys;
    }

    public Map<String, String> getTablesType() {
        return tablesType;
    }
}

class ColumnsInfo implements Serializable {

    private static final long serialVersionUID = -3407052226160044192L;

    private final Map<String, Map<String, List<ColumnInfo>>> columnsInfo;

    ColumnsInfo() {
        this.columnsInfo = new HashMap<String, Map<String, List<ColumnInfo>>>();
    }

    void addColumnsInfo(Connection sourceConnection, TableEntity tableEntity) throws SQLException {
        String schemaName = tableEntity.getSchemaName();
        Map<String, List<ColumnInfo>> map = columnsInfo.get(schemaName);
        if (map == null) {
            map = new HashMap<String, List<ColumnInfo>>();
            columnsInfo.put(schemaName, map);
        }
        map.put(tableEntity.getShortTableName(), DataProvider.getAllColumns(sourceConnection, tableEntity));
    }

    List<ColumnInfo> getColumnsInfo(String schemaName, String shortTableName) {
        return columnsInfo.get(schemaName).get(shortTableName);
    }
}

/**
 * This class manage load of TAP_SCHEMA entities and other database schemas
 * informations. IMPORTANT: If you use this class inside another object remember
 * to call the close method to release the occupied resources.
 *
 * @author Sonia Zorba <zorba at oats.inaf.it>
 */
public class TapSchemaHandler implements Serializable, Closeable {

    private static final long serialVersionUID = 1193743698810532785L;
    private static final Logger log = LoggerFactory.getLogger(TapSchemaHandler.class);

    private final Credentials sourceCredentials;
    private final Credentials tapSchemaCredentials;
    private final String name;
    private boolean exists;

    private final List<String> allSchemas;
    private final List<SchemaEntity> schemas;
    private final List<String> persistedSchemas;
    private final List<SchemaEntity> toRemoveSchemas;

    private final Map<String, SchemaInfo> allSchemaInfos;

    private final ColumnsInfo columnsInfo;

    private transient EntityManager entityManager;

    public TapSchemaHandler(Credentials sourceCredentials, Credentials tapSchemaCredentials, String tapSchemaName, boolean exists) throws SQLException {
        this.sourceCredentials = sourceCredentials;
        this.tapSchemaCredentials = tapSchemaCredentials;
        this.name = tapSchemaName;
        this.exists = exists;

        Connection sourceConn = sourceCredentials.getConnection();

        try {
            this.allSchemas = DataProvider.getAllSchemasNames(sourceConn);

            this.schemas = new ArrayList<SchemaEntity>();
            this.persistedSchemas = new ArrayList<String>();
            this.toRemoveSchemas = new ArrayList<SchemaEntity>();

            this.allSchemaInfos = new HashMap<String, SchemaInfo>();

            this.columnsInfo = new ColumnsInfo();

            if (exists) {
                loadEntityManager();
                CriteriaBuilder cb = entityManager.getCriteriaBuilder();
                CriteriaQuery<SchemaEntity> cq = cb.createQuery(SchemaEntity.class);
                // Load all SchemaEntities saved into the TAP_SCHEMA
                for (SchemaEntity schemaEntity : entityManager.createQuery(cq).getResultList()) {
                    String schemaName = schemaEntity.getName();
                    addSchema(schemaEntity);

                    Connection conn = name.equals(schemaName) ? tapSchemaCredentials.getConnection() : sourceConn;

                    SchemaInfo schemaInfo = new SchemaInfo(conn, schemaName);
                    allSchemaInfos.put(schemaName, schemaInfo);

                    persistedSchemas.add(schemaName);
                    for (TableEntity tableEntity : schemaEntity.getTables()) {
                        columnsInfo.addColumnsInfo(conn, tableEntity);
                    }

                    if (name.equals(schemaName)) {
                        closeConnection(conn);
                    }
                }
            }
        } catch (SQLException e) {
            throw e;
        } finally {
            closeConnection(sourceConn);
        }
    }

    public SchemaEntity getNewSchema(String schemaName) throws SQLException {
        return getNewSchema(schemaName, sourceCredentials.getConnection(), true);
    }

    public SchemaEntity getNewSchema(String schemaName, Connection conn, boolean closeConnection) throws SQLException {
        SchemaEntity schemaEntity = new SchemaEntity(schemaName);

        try {
            allSchemaInfos.put(schemaName, new SchemaInfo(conn, schemaName));
        } catch (SQLException e) {
            throw e;
        } finally {
            if (closeConnection) {
                closeConnection(conn);
            }
        }

        return schemaEntity;
    }

    public final void addSchema(SchemaEntity schemaEntity) {
        DLUtil.insertEntityInOrder(schemas, schemaEntity);
    }

    public final void removeSchema(String schemaName) {
        if (persistedSchemas.contains(schemaName)) {
            toRemoveSchemas.add(getSchema(schemaName));
            persistedSchemas.remove(schemaName);
        }
        DLUtil.removeEntity(schemas, schemaName);
    }

    public List<SchemaEntity> getSchemas() {
        return schemas;
    }

    public List<String> getAllSchemas() {
        return allSchemas;
    }

    public List<String> getAllTables(String schemaName) {
        return allSchemaInfos.get(schemaName).getAllSchemaTables();
    }

    public List<ColumnInfo> getColumnInfo(String schemaName, String shortTableName) {
        return columnsInfo.getColumnsInfo(schemaName, shortTableName);
    }

    public int getMaxKey() {
        int maxKey = 0;
        for (SchemaEntity schemaEntity : schemas) {
            for (TableEntity tableEntity : schemaEntity.getTables()) {
                for (KeyEntity keyEntity : tableEntity.getFromKeys()) {
                    int key = Integer.parseInt(keyEntity.getKeyId());
                    if (key > maxKey) {
                        maxKey = key;
                    }
                }
                for (KeyEntity keyEntity : tableEntity.getTargetKeys()) {
                    int key = Integer.parseInt(keyEntity.getKeyId());
                    if (key > maxKey) {
                        maxKey = key;
                    }
                }
            }
        }
        return maxKey;
    }

    public TableEntity getNewTable(String schemaName, String shortTableName) throws SQLException {
        return getNewTable(schemaName, shortTableName, sourceCredentials.getConnection(), true);
    }

    public TableEntity getNewTable(String schemaName, String shortTableName, Connection conn, boolean closeConnection) throws SQLException {
        TableEntity table = new TableEntity(getSchema(schemaName), shortTableName);

        table.setTableType(allSchemaInfos.get(schemaName).getTablesType().get(shortTableName));

        try {
            // Add columns info
            // For performance reasons all possible column entities are loaded together
            columnsInfo.addColumnsInfo(conn, table);
        } catch (SQLException e) {
            throw e;
        } finally {
            if (closeConnection) {
                closeConnection(conn);
            }
        }

        return table;
    }

    public void addTable(String schemaName, TableEntity tableEntity) {
        SchemaEntity schema = getSchema(schemaName);
        schema.addTable(tableEntity);

        // Add table keys
        int currentKey = getMaxKey() + 1;
        for (KeyEntity keyEntity : getAllSchemaKeys(schemaName)) {
            String fromTableName = keyEntity.getFromTableFullName();
            String targetTableName = keyEntity.getTargetTableFullName();
            TableEntity fromTable = null, targetTable = null;
            if (tableEntity.getName().equals(fromTableName)) {
                fromTable = tableEntity;
                targetTable = schema.getTable(targetTableName);
            } else if (tableEntity.getName().equals(targetTableName)) {
                targetTable = tableEntity;
                fromTable = schema.getTable(fromTableName);
            }
            if (fromTable != null && targetTable != null
                    && !fromTable.getFromKeys().contains(keyEntity)
                    && !targetTable.getTargetKeys().contains(keyEntity)) {
                String key = currentKey + "";
                keyEntity.setKeyId(key);
                keyEntity.setFromTable(fromTable);
                keyEntity.setTargetTable(targetTable);
                for (KeyColumnEntity keyColumn : keyEntity.getKeyColumns()) {
                    keyColumn.setKeyId(key);
                }

                fromTable.addFromKey(keyEntity);
                targetTable.addTargetKey(keyEntity);

                currentKey++;
            }
        }
    }

    public void removeTable(SchemaEntity schemaEntity, TableEntity tableEntity) {
        for (KeyEntity keyEntity : tableEntity.getFromKeys()) {
            keyEntity.getTargetTable().removeTargetKey(keyEntity);
        }
        for (KeyEntity keyEntity : tableEntity.getTargetKeys()) {
            keyEntity.getFromTable().removeFromKey(keyEntity);
        }

        schemaEntity.removeTable(tableEntity.getName());
    }

    public void addColumn(TableEntity tableEntity, ColumnEntity columnEntity) {
        tableEntity.addColumn(columnEntity);
    }

    public void removeColumn(TableEntity tableEntity, String columnName) {
        tableEntity.removeColumn(columnName);
    }

    public String getForeignKeyReference(String schemaName, String fullTableName, String columnName) {
        for (KeyEntity key : getAllSchemaKeys(schemaName)) {
            if (key.getFromTableFullName().equals(fullTableName)) {
                for (KeyColumnEntity keyColumn : key.getKeyColumns()) {
                    if (keyColumn.getFromColumn().equals(columnName)) {
                        return key.getTargetTableFullName().split(Pattern.quote("."))[1] + "." + keyColumn.getTargetColumn();
                    }
                }
            }
        }

        return null;
    }

    public SchemaEntity getSchema(String schemaName) {
        return DLUtil.getEntityByName(schemas, schemaName);
    }

//    private Connection getSourceConnection() throws SQLException {
//        if (connection == null) {
//            connection = sourceCredentials.getConnection();
//        }
//        return connection;
//    }
//
//    private Connection getTapSchemaConnection() throws SQLException {
//        if (connection == null) {
//            connection = tapSchemaCredentials.getConnection();
//        }
//        return connection;
//    }
    private void closeConnection(Connection connection) throws SQLException {
        if (connection != null) {
            try {
                connection.close();
            } catch (SQLException e) {
                log.error("Can't close connection", e);
            }
        }
    }

    public List<KeyEntity> getAllSchemaKeys(String schemaName) {
        return allSchemaInfos.get(schemaName).getSchemaKeys();
    }

    public void save() throws SQLException {
        if (exists) {
            loadEntityManager();

            entityManager.getTransaction().begin();
            for (SchemaEntity toRemoveSchema : toRemoveSchemas) {
                entityManager.remove(toRemoveSchema);
            }
            for (SchemaEntity schemaEntity : schemas) {
                entityManager.merge(schemaEntity);
            }
            entityManager.getTransaction().commit();
        } else {
            Connection tapSchemaConn = null;
            try {
                tapSchemaConn = tapSchemaCredentials.getConnection();
                DataProvider.createDatabase(tapSchemaConn, name);
            } catch (SQLException e) {
                closeConnection(tapSchemaConn);
                throw e;
            }

            loadEntityManager();

            entityManager.getTransaction().begin();
            for (SchemaEntity schemaEntity : schemas) {
                entityManager.persist(schemaEntity);
            }
            entityManager.getTransaction().commit();

            // Adding TAP_SCHEMA into TAP_SCHEMA
            log.info("Adding TAP_SCHEMA into TAP_SCHEMA");
            SchemaEntity tapSchemaSchema;
            try {
                tapSchemaSchema = getNewSchema(name, tapSchemaConn, false);
                addSchema(tapSchemaSchema);

                for (String shortTableName : allSchemaInfos.get(name).getAllSchemaTables()) {
                    TableEntity tableEntity = getNewTable(name, shortTableName, tapSchemaConn, false);
                    addTable(name, tableEntity);
                    for (ColumnInfo ci : columnsInfo.getColumnsInfo(name, shortTableName)) {
                        addColumn(tableEntity, ci.getColumnEntity());
                    }
                }
            } catch (SQLException e) {
                log.error("Exception", e);
                throw e;
            } finally {
                closeConnection(tapSchemaConn);
            }

            DLUtil.putInfoIntoTapSchemaSchema(tapSchemaSchema);

            entityManager.getTransaction().begin();
            entityManager.merge(tapSchemaSchema);
            entityManager.getTransaction().commit();

            exists = true;
        }
    }

    private void loadEntityManager() {
        if (entityManager == null) {
            entityManager = DataProvider.getEntityManager(tapSchemaCredentials, name);
        }
    }

    @Override
    public void close() {

        if (entityManager != null && entityManager.isOpen()) {
            EntityManagerFactory efFactory = entityManager.getEntityManagerFactory();

            entityManager.close();

            if (efFactory.isOpen()) {
                efFactory.close();
            }
        }
    }

    /**
     * Automatic closing on serialization
     *
     * @param stream
     * @throws IOException
     */
    private void writeObject(ObjectOutputStream stream) throws IOException {
        close();
        stream.defaultWriteObject();
    }
}
